# -*- encoding: utf-8 -*-
##
##
## Copyright (C) 2011 Thibaut Girka <thib@sitedethib.com>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation; version 3 only.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##

import weakref

from fyui.vm.boss import BossRunner

from fyui.game.gamestate import GameState
from fyui.game.sprite import Sprite
from fyui.game.bullet import Bullet

from math import cos, sin, atan2, pi, log


class Enemy(Sprite):
    def __init__(self, game, texcoords, hitbox, life=20):
        Sprite.__init__(self, texcoords, hitbox=hitbox)
        self.game = game
        self.bullet_types = []

        self.speed_modifier = 1.0
        self.rotation_speed = 0.01
        self.life = life
        self.frame = 0


    def on_hit(self):
        self.life -= 1


    def fire(self, mode, set_type, bullet_type_idx, number, param):
        set_type %= 4
        bullet_type_idx %= len(self.bullet_types)
        player = self.game.player
        angle = 0.
        if set_type == 0: # Straight toward player
            number = (number % 6) + 1 # Ensure at least 1 bullet
            speed = 2.6
            launch_angle = atan2(player.y - self.y, player.x - self.x)
            param = 0.2 + (pi / 2. - 0.2) * param
            param = number * max(param / number, pi/42.)
            if number > 1:
                launch_angle -= param / 2.
                angle = param / (number - 1)
        elif set_type == 1: # Distributed equally on the whole circle, rotating if autofire
            number = (number % 15) + 5 # Ensure at least 1 bullet
            launch_angle = 0. if mode == 0. else self.frame * self.rotation_speed
            if number > 1:
                angle = pi * 2 / (number - 1)
            speed = 2. + param * 1.8
        elif set_type == 2:
            number = (number % 10) + 1
        elif set_type == 3: # Distributed equally on the whole circle, rotating
            number = (number % 10) + 4 # Ensure at least 1 bullet
            self.rotation_speed = (0.1 + param) / 20.
            launch_angle = self.frame * self.rotation_speed
            if number > 1:
                angle = pi * 2. / (number - 1)
            speed = 2.6

        for i in xrange(number):
            if set_type == 2:
                launch_angle = self.game.prng2.random() * 2. * pi
                speed = 2. + param * 3.2 * self.game.prng2.random()
            bullet = Bullet((self.x, self.y),
                            self.bullet_types[bullet_type_idx],
                            launch_angle, speed * self.speed_modifier)
            self.game.bullets.append(bullet)
            launch_angle += angle



class Drone(Enemy):
    def __init__(self, game, src, dest):
        Enemy.__init__(self, game, (224., 255-32., 32., 32.), hitbox=(-16., -16., 16., 16.))
        self.life = 10
        self.speed_modifier = 1.0
        self.bullet_types = game.bullet_types[1]
        self.autofire = (0, 0, 3, 0.4)
        self.src = src
        self.dest = dest
        self.frame = 0


    def set_disappearing(self):
        self.frame = 0
        def new_update():
            self.frame += 1
            if self.frame < 60:
                self.scale_factor += .04
                self.alpha -= 255 // 60
                return True
            return False
        self.update = new_update


    def update(self):
        if self.frame <= 120:
            (x1, y1), (x2, y2) = self.src, self.dest
            self.x = x1 + (x2 - x1) * self.frame / 120. + (self.game.prng2.random() - .5) * 3.
            self.y = y1 + (y2 - y1) * self.frame / 120. + (self.game.prng2.random() - .5) * 3.
        if self.frame == 120:
            self.update = self.update2
        self.frame += 1
        return True


    def update2(self):
        if self.life <= 0:
            self.set_disappearing()
            return True
        self.x = self.dest[0] + (self.game.prng2.random() - .5) * 3.
        self.y = self.dest[1] + (self.game.prng2.random() - .5) * 3.
        if self.frame % 22 == 0 and not self.game.boss.regenerating:
            self.fire(1, *self.autofire)
        self.frame += 1
        return True



class Boss(Enemy):
    def __init__(self, game):
        Enemy.__init__(self, game, (0., 0., 85., 126.), hitbox=(-19, -64, 19, 20))
        self.bullet_types = game.bullet_types[0]
        self.vm = BossRunner(weakref.proxy(self))

        self.x, self.y = 200.0, 400.0
        self.frame = 0
        self.regenerating = False
        self.regeneration_start = 0
        self.exec_frame = 0
        self.exec_pointer = 0

        self.dead = False
        self.life = 42
        self.maxlife = self.life
        self.autofire = (0, 0, 3, 0.4)

        self.data = [[0, 0, 0, 0, 0] for i in xrange(30)]
        for i in xrange(30): # Pre-mutate a bit
            self.mutate()
        self.data.extend([0, 0, 0, 0, 0] for i in xrange(5))

        self.script = []
        self.effects = []
        self.wings = []
        self.set_wings((85, 0, 72, 45))

        self.regen(False)


    def set_wings(self, texcoords):
        self.wings = [Sprite(texcoords, True, offset=(72, -45)),
                      Sprite(texcoords, True, offset=(-72, -45))]
        self.wings[1].flip()


    def mutate(self):
        prng = self.game.prng
        prng.choice(self.data)[prng.randrange(4)] ^= (1 << prng.randrange(8))


    def regen(self, dead=True):
        if dead:
            self.game.score += 1000 + 100 * self.game.nb_takedowns
            self.game.nb_takedowns += 1
            self.regenerating = True
            self.regeneration_start = self.frame
            self.effects = [Sprite((32, 256-64, 64, 64), blend_mode=1, pos=(self.x, self.y)),
                            Sprite((0, 256-32, 32, 32), blend_mode=1, pos=(self.x, self.y)),
                            Sprite((0, 256-32, 32, 32), blend_mode=1, pos=(self.x, self.y)),
                            Sprite((0, 256-32, 32, 32), blend_mode=1, pos=(self.x, self.y)),
                            Sprite((0, 256-32, 32, 32), blend_mode=1, pos=(self.x, self.y)),
                            Sprite((0, 256-32, 32, 32), blend_mode=1, pos=(self.x, self.y))]
        self.dead = not self.vm.reload_script() and not self.game.survival_mode
        self.data.append([0, 0, 0, 0, 0])
        self.life = 40 + min(self.life, 120)
        self.maxlife = self.life


    def on_hit(self):
        if self.regenerating:
            return
        else:
            self.mutate()
            self.life -= 1
            if self.life <= 0:
                self.regen()


    def change_bullet_properties(self, bullet_type_idx, accel, rotation_speed):
        bullet_type_idx %= len(self.bullet_types)
        bullet_type = self.bullet_types[bullet_type_idx]
        for bullet in self.game.bullets:
            if bullet.bullet_type is bullet_type:
                bullet.accel = 0.06 * (accel - 0.5)
                bullet.rotation_speed = 0.02 * (rotation_speed - 0.5)


    def update(self):
        if self.frame < 180:
            self.y = 700 + (380-700) * self.frame / 180.
        else:
            self.y = 380.0 - 5. * sin((self.frame - 180.) / 20.)

        self.wings[0].x = self.x - 10
        self.wings[0].y = self.y + 8

        self.wings[1].x = self.x + 10
        self.wings[1].y = self.y + 8

        self.wings[0].angle = -pi / 2. + pi / 10. * sin(2 * pi * self.frame / 10.) #TODO
        self.wings[1].angle = -pi / 2. - pi / 10. * sin(2 * pi * self.frame / 10.) #TODO

        if self.frame >= 240 and not self.regenerating:
            script = self.script
            while script:
                try:
                    frame, instructions = script[self.exec_pointer]
                except IndexError:
                    last_frame = script[-1][0] if script else 0
                    if self.exec_frame > last_frame + 120:
                        self.exec_frame = 0
                        self.exec_pointer = 0
                    break
                if frame > self.exec_frame:
                    break
                self.exec_pointer += 1
                for opcode, args in instructions:
                    if opcode == 1:
                        self.autofire = args
                    elif opcode == 2:
                        self.fire(0, *args)
                    elif opcode == 3:
                        self.change_bullet_properties(*args)
                    elif opcode == 4:
                        self.speed_modifier = 1.1 + args[0] * 1.4
                    elif opcode == 5: # Drones \o/
                        drone_id = args[0]
                        if not self.game.drones[drone_id]:
                            pos = (self.x, self.y)
                            if drone_id == 0:
                                dest = (80, self.game.height - 80)
                            else:
                                dest = (self.game.width - 80, self.game.height - 80)
                            self.game.drones[drone_id] = Drone(self.game, pos, dest)
                        drone = self.game.drones[drone_id]
                        drone.speed_modifier = min(2., drone.speed_modifier + .02)
                        drone.autofire = args[1:]

            if self.frame % 20 == 0:
                self.fire(1, *self.autofire)

            self.exec_frame += 1
        elif self.regenerating:
            ellapsed_time = self.frame - self.regeneration_start
            master_effect = self.effects[0]

            master_effect.scale_factor = log(ellapsed_time + 1, 60) * 4.
            for i, effect in enumerate(self.effects[1:]):
                angle = ellapsed_time * pi / 30. + i*2.*pi/5.
                effect.x = master_effect.x + cos(angle) * ellapsed_time * 2.5
                effect.y = master_effect.y + sin(angle) * ellapsed_time * 2.5

            if not self.dead:
                if ellapsed_time == 110:
                    self.effects = []
                    self.regenerating = False
                elif ellapsed_time > 90:
                    master_effect.scale_factor += (self.frame - self.regeneration_start - 90) / 2.
                    for effect in self.effects:
                        effect.alpha -= 12
            else:
                if ellapsed_time == 180:
                    self.effects = []
                    self.regenerating = False
                elif ellapsed_time > 110:
                    master_effect.alpha -= 255 // 70
                    master_effect.scale_factor += (self.frame - self.regeneration_start - 90) / 2.
                    self.game.state = GameState.WON
                elif ellapsed_time > 90:
                    master_effect.scale_factor += (self.frame - self.regeneration_start - 90) / 2.

        self.frame += 1

