# -*- encoding: utf-8 -*-
##
##
## Copyright (C) 2011 Thibaut Girka <thib@sitedethib.com>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation; version 3 only.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##

import os

from fyui.game.replay import Replay
from fyui.game.gamestate import GameState
from fyui.screens.scoresscreen import ScoresScreen
from fyui.screens.savereplayscreen import SaveReplayScreen

from math import cos, sin, pi

from ctypes import byref, c_buffer, c_uint
from struct import Struct

from itertools import chain, groupby

import pyglet

from pyglet.gl import *

from operator import attrgetter

MAX_ELEMENTS = 10000

class TextureManager(dict):
    def load_texture(self, name):
        image = pyglet.resource.image(name)
        return image.get_texture()

    def __missing__(self, key):
        value = self.load_texture(key)
        self[key] = value
        glTexParameteri(value.target, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        glTexParameteri(value.target, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        return value


class GameOverScreen(object):
    def __init__(self, screen_list):
        self.screen_list = screen_list

        self.background = pyglet.resource.image('game_over.png')


    def on_key_press(self, key, modifiers):
        if key in (pyglet.window.key.ESCAPE, pyglet.window.key.ENTER, pyglet.window.key.LCTRL):
            self.screen_list.remove(self)


    def update(self, keystate):
        pass


    def render(self):
        self.background.blit(0, 0)



class GameScreen(object):
    def __init__(self, game, screen_list):
        self.screen_list = screen_list

        self.game = game
        self.replay = Replay(game.seed, game.seed2, survival_mode=game.survival_mode)

        self.paused = False
        self.texture_manager = TextureManager()

        self.score_label = pyglet.text.Label('0', font_name='Mono', font_size=40, anchor_x='right', anchor_y='center')
        self.lives_label = pyglet.text.Label('0', font_name='Mono', font_size=40, anchor_x='center', anchor_y='center')
        self.bombs_label = pyglet.text.Label('0', font_name='Mono', font_size=40, anchor_x='center', anchor_y='center')

        # Generate texture for off-screen rendering
        self.fbo_texture = pyglet.image.Texture.create(self.game.width, self.game.height, rectangle=True)
        buffer_id = c_uint()
        glGenFramebuffersEXT(1, byref(buffer_id))
        self.fbo = buffer_id.value
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, self.fbo)
        glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT,
                                                      self.fbo_texture.target,
                                                      self.fbo_texture.id, 0)
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0)

        # Allocate Vertex+TexCoord+Color buffer
        buff = c_buffer(MAX_ELEMENTS * (4*4*2+4*4*2+4*4))
        self.buffer = buff

        self.frame = 0


    def on_key_press(self, key, modifiers):
        if key == pyglet.window.key.SPACE:
            self.paused = not self.paused
        elif key == pyglet.window.key.ESCAPE:
            self.screen_list.remove(self)


    def render(self):
        glPushAttrib(GL_ALL_ATTRIB_BITS)
        glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT)
        self.setup_opengl()

        # Render
        self.render_ui()
        self.render_game()

        if self.frame <= 60:
            glViewport(0, 0, 520, 500)
            glDisable(GL_SCISSOR_TEST)
            glDisable(GL_TEXTURE_2D)
            glBegin(GL_QUADS)
            glColor4f(0.0, 0.0, 0.0, (60.-self.frame)/60.)
            glVertex2f(0, 0)
            glColor4f(0.0, 0.0, 0.0, (60.-self.frame)/60.)
            glVertex2f(0, 500)
            glColor4f(0.0, 0.0, 0.0, (60.-self.frame)/60.)
            glVertex2f(520, 500)
            glColor4f(0.0, 0.0, 0.0, (60.-self.frame)/60.)
            glVertex2f(520, 0)
            glEnd()

        # Pop attribs
        glPopClientAttrib()
        glPopAttrib()


    def setup_opengl(self):
        glVertexPointer(2, GL_FLOAT, 20, self.buffer)
        glTexCoordPointer(2, GL_FLOAT, 20, byref(self.buffer, 4*2))
        glColorPointer(4, GL_UNSIGNED_BYTE, 20, byref(self.buffer, 4*2+4*2))

        glEnable(GL_BLEND)
        glEnable(GL_TEXTURE_2D)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST)
        glEnableClientState(GL_VERTEX_ARRAY)
        glEnableClientState(GL_TEXTURE_COORD_ARRAY)
        glEnableClientState(GL_COLOR_ARRAY)


    def update(self, keystate):
        if self.game.state != GameState.PLAYING:
            self.replay.score = self.game.score
            self.screen_list.remove(self)
            self.screen_list.append(SaveReplayScreen(self.screen_list, self.replay))
            self.screen_list.append(ScoresScreen(self.screen_list, self.game))
            if self.game.state == GameState.LOST:
                self.screen_list.append(GameOverScreen(self.screen_list))
            return
        if not self.paused:
            self.replay.add_keystate(self.frame, keystate)
            self.game.update(keystate)
            self.frame += 1


    def render_hitboxes(self):
        sprites = chain(self.game.bullets,
                        self.game.player_bullets,
                        (self.game.player,))

        pack_hitbox = Struct('ff 8x 4B' * 4).pack_into
        buff = self.buffer

        for i, sprite in enumerate(sprites):
            if not sprite.hitbox:
                continue
            else:
                ox, oy = sprite.x, sprite.y
                x1, y1, x2, y2 = sprite.hitbox
                pack_hitbox(buff, i * 20 * 4,
                                        ox + x1, oy + y1, 255, 255, 255, sprite.alpha,
                                        ox + x2, oy + y1, 255, 255, 255, sprite.alpha,
                                        ox + x2, oy + y2, 255, 255, 255, sprite.alpha,
                                        ox + x1, oy + y2, 255, 255, 255, sprite.alpha)

        # Render in off-screen FBO
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, self.fbo)
        # Set it up
        glPushAttrib(GL_SCISSOR_BIT | GL_VIEWPORT_BIT | GL_TEXTURE_BIT)
        glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT)
        glDisable(GL_SCISSOR_TEST)
        glDisableClientState(GL_TEXTURE_COORD_ARRAY)
        glDisable(GL_TEXTURE_2D)

        glViewport(0, 0, self.game.width+120, self.game.height)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        # Clear it
        glClearColor(0.0, 0.0, 0.0, 0.0)
        glClear(GL_COLOR_BUFFER_BIT)

        glDrawArrays(GL_QUADS, 0, (i + 1) * 4)

        # Restore settings
        glPopClientAttrib()
        glPopAttrib()

        # Render on_screen_fbo
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0)
        glBlendFunc(GL_ONE_MINUS_DST_COLOR, GL_ONE_MINUS_SRC_COLOR)
        self.fbo_texture.blit(0, 0)


    def render_sprites(self, sprites):
        pack_sprite = Struct('ff ff BBBB' * 4).pack_into
        buff = self.buffer

        sprites = sorted(filter(None, sprites), key=attrgetter('blend_mode'))
        for blend_mode, sprites in groupby(sprites, key=attrgetter('blend_mode')):
            for i, sprite in enumerate(sprites):
                ox, oy = sprite.x, sprite.y

                (x1, x2, x3, x4), (y1, y2, y3, y4) = sprite.vertmat
                u1, v1, u2, v2, u3, v3, u4, v4 = sprite.texcoords

                if sprite.center_offset is not None:
                    # Hackish, but well
                    cox, coy = sprite.center_offset
                    cox *= .5
                    coy *= .5
                    x1, x2, x3, x4 = x1 - cox, x2 - cox, x3 - cox, x4 - cox
                    y1, y2, y3, y4 = y1 - coy, y2 - coy, y3 - coy, y4 - coy

                if sprite.scale_factor != 1.:
                    s = sprite.scale_factor
                    x1, x2, x3, x4 = x1 * s, x2 * s, x3 * s, x4 * s
                    y1, y2, y3, y4 = y1 * s, y2 * s, y3 * s, y4 * s

                if sprite.automatic_rotation:
                    angle = sprite.angle + pi / 2.
                    cos_a, sin_a = cos(angle), sin(angle)
                    (x1, x2, x3, x4), (y1, y2, y3, y4) = ((cos_a * x1 - sin_a * y1,
                                                           cos_a * x2 - sin_a * y2,
                                                           cos_a * x3 - sin_a * y3,
                                                           cos_a * x4 - sin_a * y4),
                                                          (sin_a * x1 + cos_a * y1,
                                                           sin_a * x2 + cos_a * y2,
                                                           sin_a * x3 + cos_a * y3,
                                                           sin_a * x4 + cos_a * y4))

                pack_sprite(buff, i * 20 * 4,
                            x1 + ox, y1 + oy, u1, v1, 255, 255, 255, sprite.alpha,
                            x2 + ox, y2 + oy, u2, v2, 255, 255, 255, sprite.alpha,
                            x3 + ox, y3 + oy, u3, v3, 255, 255, 255, sprite.alpha,
                            x4 + ox, y4 + oy, u4, v4, 255, 255, 255, sprite.alpha)
            glVertexPointer(2, GL_FLOAT, 20, self.buffer)
            glTexCoordPointer(2, GL_FLOAT, 20, byref(self.buffer, 4*2))
            glColorPointer(4, GL_UNSIGNED_BYTE, 20, byref(self.buffer, 4*2+4*2))
            glBlendFunc(GL_SRC_ALPHA, [GL_ONE_MINUS_SRC_ALPHA, GL_ONE][blend_mode])
            glDrawArrays(GL_QUADS, 0, (i + 1) * 4)


    def render_ui(self):
        pack_sprite = Struct('ff 8x BBBB' * 4).pack_into
        buff = self.buffer

        glDisable(GL_SCISSOR_TEST)
        glViewport(0, 0, self.game.width+120, self.game.height)

        # Clear
        glClearColor(0, 0, 0, 1.0)
        glClear(GL_COLOR_BUFFER_BIT)

        # Draw score
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()
        glTranslatef(30, self.game.height, 0)
        glRotatef(-90, 0.0, 0.0, -1.0)
        self.score_label.text = '%010d' % self.game.score
        self.score_label.draw()

        # Draw lives
        glLoadIdentity()
        glTranslatef(self.game.width + 60 + 30, self.game.height - 60, 0)
        self.lives_label.text = '%d' % max(0, self.game.player.lives)
        self.lives_label.draw()

        # Draw bombs
        glLoadIdentity()
        glTranslatef(self.game.width + 60 + 30, 60, 0)
        glPushAttrib(GL_TEXTURE_BIT)
        glDisable(GL_TEXTURE_2D)
        next_bomb_completion = self.game.player.next_bomb_completion
        pack_sprite(buff, 0,
                    -20, -40, 127, 127, 255, 255,
                    -20,  -40 + 80. * next_bomb_completion, 127, 127, 255, 255,
                     20,  -40 + 80. * next_bomb_completion, 127, 127, 255, 255,
                     20, -40, 127, 127, 255, 255)
        glDrawArrays(GL_QUADS, 0, 4)
        pack_sprite(buff, 0,
                    -20, -40, 255, 255, 255, 255,
                    -20,  40, 255, 255, 255, 255,
                     20,  40, 255, 255, 255, 255,
                     20, -40, 255, 255, 255, 255)
        glDrawArrays(GL_LINE_LOOP, 0, 4)
        glPopAttrib()
        self.bombs_label.text = '%d' % self.game.player.bombs
        self.bombs_label.draw()


    def render_game(self):
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

        glEnable(GL_SCISSOR_TEST)
        glViewport(60, 0, self.game.width+120, self.game.height)
        glScissor(60, 0, self.game.width, self.game.height)

        # Local lookup is faster than global
        game = self.game

        # Clear
        glClearColor(0.3, 0.3, 0.3, 1.0)
        glClear(GL_COLOR_BUFFER_BIT)

        # Render boss
        texture = self.texture_manager['boss.png']
        glBindTexture(texture.target, texture.id)
        self.render_sprites(game.drones)
        self.render_sprites(game.boss.wings)
        self.render_sprites((game.boss,))

        # Render player bullets
        texture = self.texture_manager['bullets.png']
        glBindTexture(texture.target, texture.id)
        self.render_sprites(game.player_bullets)

        # Render player
        texture = self.texture_manager['player.png']
        glBindTexture(texture.target, texture.id)
        self.render_sprites((game.player,))

        # Render enemy bullets
        texture = self.texture_manager['bullets.png']
        glBindTexture(texture.target, texture.id)
        self.render_sprites(chain(game.bullets, game.pending_deletion))

        # Render effects
        texture = self.texture_manager['effects.png']
        glBindTexture(texture.target, texture.id)
        self.render_sprites(chain(game.boss.effects,
                                  game.player.effects,
                                  game.player.bomb_particles))

        # Render hitboxes
        if game.player.slow_mode:
            self.render_hitboxes()

